#include <stdio.h>
#include <usb.h>

#include "defs.h"

#ifndef LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
#warning : usb_detach_kernel_driver_np not available, you must do "rmmod usbhid" manually!
#endif
#ifndef LIBUSB_HAS_GET_DRIVER_NP
#warning : usb_detach_get_driver_np not available, you must do "rmmod usbhid" manually!
#endif

/* vendor and device IDs */
#define SYNAPTICS_VENDOR_ID	0x06cb
static struct {
	int dev_id;
	char *dev_name;
} dev_names[] = {
	{ 0x0001, "Synaptics USB TouchPad" },
	{ 0x0002, "Synaptics Integrated USB TouchPad" },
	{ 0x0003, "Synaptics cPad" },
	{ 0x0006, "Synaptics TouchScreen" },
	{ 0x0007, "Synaptics USB Styk" },
	{ 0x0008, "Synaptics USB WheelPad" },
	{ 0x0009, "Synaptics Composite USB Device" },
	{ 0, NULL }
};

static struct usb_dev_handle *usb_handle = NULL;
static int endpoint;
static unsigned char* data = NULL;
static int max_data;
static int claimed = 0;

static int int_num = ALLOW_TP ? 0 : 1;

static enum data_type {pad, stick, screen} type;

static inline int us13_to_int(unsigned char high, unsigned char low, int has_sign) {
	int res;

	res = ((int)(high & 0x1f) << 8) | low;
	if (has_sign && (high & 0x10))
		res -= 0x2000;

	return res;
}

#define XMIN 1472
#define XMAX 5472
#define YMIN 1408
#define YMAX 4448

static int decode_data(struct input_data *in)
{
	int data_size;
	int X, Y, B, P, W, touched;

	data_size = usb_interrupt_read(usb_handle, endpoint, (char *) data, max_data, 0);
	if (data_size < 0) {
		perror("can not read from endpoint");
		return -1;
	}

	switch (type) {
	case screen:
		/* FIXME */
	case pad:
		X = us13_to_int(data[2], data[3], 0);
		Y = us13_to_int(data[4], data[5], 0);
		B = data[1];
		P = data[6];
		W = data[0] & 0x0f;
		Y = YMIN + YMAX - Y;
		break;
	case stick:
		X = us13_to_int(data[2], data[3], 1);
		Y = us13_to_int(data[4], data[5], 1);
		B = data[1];
		P = data[6];
		break;
	}

	if (P > 6)
		touched = 1;
	if (P < 5)
		touched = 0;
	in->touched = touched;

	in->X = X;
	in->Y = Y;
	in->P = P;
	in->buttons.left = (B & 0x04) ? 1 : 0;
	in->buttons.right = (B & 0x01) ? 1 : 0;
	in->buttons.middle = (B & 0x02) ? 1 : 0;

	return 0;
}

static void cleanup(void)
{
	if (data)
		free(data);
	if (claimed)
		if (usb_release_interface(usb_handle, int_num) != 0)
			perror("can not release inertface");
	if (usb_handle != NULL)
		if (usb_close(usb_handle) != 0)
			perror("can not close interface");
}

static char* get_dev_name(struct usb_device *dev)
{
	int i;

	for (i=0; dev_names[i].dev_name; i++) {
		if (dev->descriptor.idProduct == dev_names[i].dev_id)
			break;
	}

	return dev_names[i].dev_name ? dev_names[i].dev_name : "unknown";
}

static void get_data_type(struct usb_device *dev)
{
	switch (dev->descriptor.idProduct) {
	case 0x0006:
		type = screen;
		break;
	case 0x0007:
		type = stick;
		break;
	case 0x0009:
		if (int_num == 1) {
			type = stick;
			break;
		}
	default:
		type = pad;
	}
}

static int find_endpoint(struct usb_interface_descriptor* interface)
{
	struct usb_endpoint_descriptor* ep;
	int i;

	for (i=0; i<interface->bNumEndpoints; i++) {
		ep = &interface->endpoint[i];
		if ( (ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) &&
		     ((ep->bmAttributes & USB_ENDPOINT_TYPE_MASK)
					== USB_ENDPOINT_TYPE_INTERRUPT) ) {
			endpoint = ep->bEndpointAddress;
			printf("found endpoint with address 0x%x\n", endpoint);
			max_data = ep->wMaxPacketSize;
			data = malloc(max_data);
			if (!data) {
				perror("can not allocate buffer");
				return -1;
			}
			return 0;
		}
	}

	fprintf(stderr, "no usable endpoint found\n");
	return -1;
}

static int claim_device(struct usb_device *dev)
{
	int num_int;
	char mod_name[16];

	usb_handle = usb_open(dev);
	if (usb_handle == NULL) {
		perror("not able to claim the USB device");
		return -1;
	}

	if (dev->descriptor.bNumConfigurations != 1) {
		fprintf(stderr, "device has wrong number of configurations: %i\n",
			dev->descriptor.bNumConfigurations);
		return -1;
	}
	num_int = dev->config[0].bNumInterfaces;
	printf("number of interfaces: %i\n", num_int);
	if (int_num >= num_int) {
		fprintf(stderr, "requested interface %i not present\n", int_num);
		return -1;
	}

#if defined(LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP) && defined(LIBUSB_HAS_GET_DRIVER_NP)
	if (usb_get_driver_np(usb_handle, int_num, mod_name, 16) == 0) {
		printf("interface %i is already attached to driver %s, detaching...\n",
			int_num, mod_name);
		if (usb_detach_kernel_driver_np(usb_handle, int_num) != 0) {
			perror("can not detach kernel driver");
			return -1;
		}
	}
#else
	printf("Driver detaching not supported by this version of libusb. You must do ""rmmod usbhid"" manually!\n");
#endif

	if (usb_claim_interface(usb_handle, int_num) != 0) {
		perror("can not claim interface");
		return -1;
	}
	claimed = 1;
	if (usb_set_altinterface(usb_handle, 1)) {
		perror("can not set alternate setting");
		return -1;
	}

	return find_endpoint(&dev->config[0].interface[int_num].altsetting[1]);
}

static int init_device(void)
{
	int res;
	struct usb_bus *usb_bus;
	struct usb_device *dev;

	usb_init();
	usb_find_busses();
	usb_find_devices();

	for (usb_bus = usb_get_busses(); usb_bus; usb_bus = usb_bus->next) {
		for (dev = usb_bus->devices; dev;dev = dev->next) {
			if ((dev->descriptor.idVendor == SYNAPTICS_VENDOR_ID)) {
				get_data_type(dev);
				if ((type == stick) || ((ALLOW_TP) && (type == pad))) {
					printf("found device: %s\n", get_dev_name(dev));
					res = claim_device(dev);
					if (res)
						return res;
					return 0;
				}
			}
		}
	}

	fprintf(stderr, "no Synaptics stick found\n");
	return -1;
}

struct input_functions input_libusb = {
	.init = init_device,
	.exit = cleanup,
	.get_data = decode_data
};
